// Microsoft Visual C++ generated resource script.
//
#include "resource.h"
#include "zstd.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "winres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// French (France) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_FRA)
LANGUAGE LANG_FRENCH, SUBLANG_FRENCH

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#include ""winres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Version
//

#define ZSTD_LIB_VERSION ZSTD_VERSION_MAJOR,ZSTD_VERSION_MINOR,ZSTD_VERSION_RELEASE,0 
#define ZRC_QUOTE(str) #str
#define ZRC_EXPAND_AND_QUOTE(str) ZRC_QUOTE(str)
#define ZSTD_LIB_VERSION_STR ZRC_EXPAND_AND_QUOTE(ZSTD_LIB_VERSION)

VS_VERSION_INFO VERSIONINFO
 FILEVERSION ZSTD_LIB_VERSION
 PRODUCTVERSION ZSTD_LIB_VERSION
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040c04b0"
        BEGIN
            VALUE "CompanyName", "Yann Collet"
            VALUE "FileDescription", "Fast and efficient compression algorithm"
            VALUE "FileVersion", ZSTD_LIB_VERSION_STR
            VALUE "InternalName", "zstdlib.dll"
            VALUE "LegalCopyright", "Copyright (C) 2013-2015, Yann Collet"
            VALUE "OriginalFilename", "zstdlib.dll"
            VALUE "ProductName", "Zstandard"
            VALUE "ProductVersion", ZSTD_LIB_VERSION_STR
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x40c, 1200
    END
END

#endif    // French (France) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

